from typing import List
from loguru import logger
from handshake.services.SchedularService.constants import (
    writtenAttachmentFolderName,
)
from shutil import rmtree, copytree
from uuid import UUID
import json
import pprint
import typing

from handshake.services.DBService.lifecycle import (
    init_tortoise_orm,
    close_connection,
    set_default_config,
    config_file,
    READ_ONLY,
    ALLOW_WRITE,
)
from pathlib import Path
from typing import Dict
from handshake.services.DBService.models.config_base import ConfigKeys, ConfigBase
from tortoise.expressions import Q


def deleteTestRunsRelatedAttachments(data_base: Path, runs: List[str]):
    collection = data_base.parent / writtenAttachmentFolderName

    for _run in runs:
        run = str(_run)  # UUID to string

        entry = collection / run
        if not entry.exists():
            continue

        logger.warning("Deleting the attachments for run {}", run)
        rmtree(entry)


def moveTestRunsRelatedAttachment(testID: UUID, copyFrom: Path, copyTo: Path):
    test_id = copyFrom / str(testID)
    logger.warning("{} to {} for {}", copyFrom, copyTo, testID)

    if not test_id.exists():
        return

    return copytree(test_id, copyTo / str(testID), dirs_exist_ok=True)


def dump_config(path: Path, to_dump: typing.List):
    formatted_dump: typing.Dict[str, str] = dict(to_dump)

    # not required
    for key in READ_ONLY:
        if key in formatted_dump:
            formatted_dump.pop(key)

    # show only required
    config_file(path).write_text(json.dumps(formatted_dump, indent=4))


async def setConfig(path: Path, feed: Dict[ConfigKeys, str], set_default: bool):
    await init_tortoise_orm(path)

    if set_default:
        await set_default_config(path)
    if feed:
        to_change = await ConfigBase.filter(Q(key__in=feed.keys())).all()
        for to in to_change:
            if to.key not in ALLOW_WRITE:
                continue

            to.value = feed[to.key]
            await ConfigBase.bulk_update(to_change, fields=["value"])

    to_dump = await ConfigBase.all().values_list()
    dump_config(path, to_dump)

    pprint.pprint(to_dump, indent=4)
    await close_connection()



import traceback
from handshake.services.DBService.models.result_base import (
    SuiteBase,
    RollupBase,
    RetriedBase,
    SessionBase,
)
from handshake.services.DBService.models.config_base import TestConfigBase
from handshake.services.DBService.models.dynamic_base import TaskBase
from handshake.services.DBService.models.enums import Status, SuiteType
from tortoise.expressions import Q
from tortoise.functions import Count, Lower, Sum
from loguru import logger
from handshake.services.SchedularService.register import skip_test_run
from itertools import chain


async def rollup_suite_values(suiteID: str):
    # rollup for the errors
    errors = list(
        chain.from_iterable(
            map(
                lambda suite_with_errors: map(
                    lambda error: error
                    | dict(
                        mailedFrom=error.get("mailedFrom", [])
                        + [str(suite_with_errors.suiteID)]
                    ),
                    suite_with_errors.errors,
                ),
                await (
                    SuiteBase.filter(Q(parent=suiteID) & ~Q(errors="[]"))
                    .only("errors", "suiteID")
                    .all()
                ),
            )
        )
    )

    suite = await SuiteBase.filter(suiteID=suiteID).first()
    await suite.update_from_dict(
        dict(
            errors=errors,
        )
    )
    await suite.save()

    required = ("passed", "failed", "skipped", "tests")

    direct_entities = (
        await (
            SuiteBase.filter(parent=suiteID, suiteType=SuiteType.TEST)
            .only("parent", *required)
            .group_by("parent")
            .annotate(**{key: Sum(key) for key in required})
            .first()
            .values(*required)
        )
        or {}
    )

    suites = await SuiteBase.filter(
        suiteType=SuiteType.SUITE, parent=suiteID
    ).values_list("suiteID", flat=True)

    indirect_entities = (
        (
            await RollupBase.filter(Q(suite_id__in=suites))
            .annotate(**{key: Sum(key) for key in required})
            .first()
            .values(*required)
        )
        if suites
        else None
    ) or {}

    await RollupBase.create(
        suite_id=suiteID,
        **{
            key: direct_entities.get(key, 0) + indirect_entities.get(key, 0)
            for key in required
        },
    )


def fetch_key_from_status(passed, failed, skipped):
    return (
        Status.FAILED
        if failed > 0
        else Status.PASSED
        if passed > 0 or skipped == 0
        else Status.SKIPPED
    )


async def handleRetries(suiteID: str, test_id: str):
    # assuming we would need to match the retries
    suite = await SuiteBase.filter(suiteID=suiteID).first()
    suite_id = str(suite.suiteID)
    if suite.retried == 0:
        return await RetriedBase.create(tests=[suite_id], suite_id=suite_id, length=1)

    matching_previous_suites = (
        await SuiteBase.filter(
            Q(title=suite.title)
            & Q(session__test_id=test_id)
            & Q(file=suite.file)
            & Q(tags=suite.tags)
            & Q(retried=suite.retried - 1)
            & Q(suiteType=suite.suiteType)
            & ~Q(suiteID=suite.suiteID)
            & Q(ended__lte=suite.started)
        )
        .order_by("started")
        .values_list("suiteID", flat=True)
    )

    previous = (
        await RetriedBase.filter(
            Q(length=suite.retried) & Q(suite_id__in=matching_previous_suites)
        )
        .order_by("modified")
        .first()
    )

    previous_suite = await SuiteBase.filter(suiteID=previous.suite_id).first()
    await previous_suite.update_from_dict(dict(standing=Status.RETRIED))
    await previous_suite.save()

    session = await SessionBase.filter(
        sessionID=(await previous_suite.session).sessionID
    ).first()
    await session.update_from_dict(dict(retried=True))
    await session.save()

    await previous.update_from_dict(
        dict(
            length=previous.length + 1,
            tests=previous.tests + [suite_id],
            suite_id=suite.suiteID,
        )
    )

    await previous.save()
    return previous


async def _patchTestSuite(suiteID: str, testID: str):
    # suiteID can also be treated as a ticketID
    task = await TaskBase.filter(ticketID=suiteID).first()
    suite = await SuiteBase.filter(suiteID=suiteID).first()
    test_config = await TestConfigBase.filter(test_id=testID).first()

    logger.info("Patching Suite: {} | {}", suite.suiteID, suite.title)

    if suite.standing != Status.YET_TO_CALCULATE:
        logger.warning("Skipping patch suite for: {}", suite.suiteID)
        return True

    pending_child_tasks = await SuiteBase.filter(
        Q(parent=suite.suiteID)
        & (Q(standing=Status.PENDING) | Q(standing=Status.YET_TO_CALCULATE))
    ).exists()

    if pending_child_tasks:
        logger.warning(
            "There are some child suites, which are not yet processed, so will process {} suite in the next iteration",
            suite.suiteID,
        )
        await task.update_from_dict(dict(picked=False))
        await task.save()  # continue in the next run
        return False

    entities = SuiteBase.filter(parent=suite.suiteID)

    results = dict(
        await (
            entities.annotate(count=Count("standing"), status=Lower("standing"))
            .group_by("standing")
            .values_list("status", "count")
        )
    )
    results["standing"] = fetch_key_from_status(
        results.get("passed", 0), results.get("failed", 0), results.get("skipped", 0)
    )
    results["tests"] = await entities.count()

    await suite.update_from_dict(results)
    await suite.save()

    await rollup_suite_values(suiteID)

    if test_config and test_config.fileRetries > 0:
        await handleRetries(suiteID, testID)

    logger.info("Successfully processed suite: {}", suite.suiteID)
    return True


async def patchTestSuite(suiteID: str, testID: str):
    try:
        return await _patchTestSuite(suiteID, testID)
    except Exception:
        await skip_test_run(
            testID,
            f"Failed to patch the test suite, found an error in calculation: {traceback.format_exc()}",
        )
        return False



import datetime
import pprint
from sqlite3 import connect
from click import (
    group,
    argument,
    secho,
    version_option,
    pass_context,
    Context,
    Path as C_Path,
)
from handshake import __version__
from handshake.services.DBService.migrator import check_version, migration, DB_VERSION
from handshake.services.SchedularService.center import start_service
from handshake.services.SchedularService.lifecycle import start_loop
from handshake.services.SchedularService.handleTestResults import (
    moveTestRunsRelatedAttachment,
    setConfig,
)
from functools import partial
from handshake.services.SchedularService.constants import writtenAttachmentFolderName
from handshake.services.SchedularService.register import (
    deleteExportTicket,
    createExportTicket,
)
from os.path import relpath
from subprocess import call, check_output
from loguru import logger
from concurrent.futures import ThreadPoolExecutor
import json
from handshake.services.DBService.lifecycle import (
    config_file,
)
from click import option
from pathlib import Path
from handshake.services.DBService.shared import db_path
from handshake.services.DBService.models.config_base import ConfigKeys
from tortoise import run_async
from os import getenv


@group(
    name="Handshake",
    short_help="Handshake command",
    help=f"""

{'{:*^69}'.format(" Welcome to Handshake's CLI ")}

Handshake simplifies the collection and processing of your test results. The Handshake CLI helps you to interact with 
the stored results or with the server. so for each command it requires a <path> argument, representing the collection 
folder, a place where you could have the database or were planning to store the results [not the HTML files].

[ROOT-DIR] >> [COLLECTION_NAME] (*we need this) >> TeStReSuLtS.db & [Attachments] (for getting this).

{'{:*^69}'.format(" Glossary ")}
""",
)
@version_option(f"{__version__}, DB: {DB_VERSION}")
def handle_cli():
    pass


general_requirement = argument(
    "COLLECTION_PATH", nargs=1, type=C_Path(exists=True, dir_okay=True), required=True
)
general_but_optional_requirement = argument(
    "COLLECTION_PATH", nargs=1, type=C_Path(dir_okay=True), required=True
)


@general_requirement
@handle_cli.command()
def db_version(collection_path):
    return check_version(db_path(collection_path))


@general_requirement
@handle_cli.command(
    short_help="Migrates the database to the latest version as per the handshake executable.",
    help="it's a command to execute the required migration scripts, note; this command would be executed "
    "automatically whenever we run patch or run-app command",
)
def migrate(collection_path: str):
    return migration(db_path(collection_path))


@handle_cli.command(
    short_help="Processes the collected results",
    help="runs an async scheduler, thanks to apscheduler, it would process your test results to patch some missing "
    "values. which are essential while showcasing your test reports",
)
@option(
    "--log-file",
    "-l",
    help="give me file name to store the logs for the patch command.",
    type=C_Path(),
    default="",
    required=False,
)
@option(
    "-r",
    "--reset",
    default=False,
    show_default=True,
    help="re-calculates the values for the test runs",
    type=bool,
    is_flag=True,
)
@general_requirement
def patch(collection_path, log_file: str, reset: bool = False):
    if log_file:
        logger.add(
            log_file if log_file.endswith(".log") else f"{log_file}.log",
            backtrace=True,
            diagnose=True,
        )

    if not Path(collection_path).is_dir():
        raise NotADirectoryError(collection_path)
    start_service(db_path(collection_path), reset)
    start_loop()


@handle_cli.command(
    help="Helper command which would assume you have nodejs installed and with the help of Next.js's SSG [Static Site "
    "Generation] we would generate reports from processed results."
    " Note: make sure to run this command from the directory where we can access the required npm scope",
    short_help="Generates the static dashboard from processed results",
)
@general_requirement
@option(
    "-mr",
    "--max-runs",
    type=int,
    default=100,
    help="Asks Sanic to set the use max. number of workers",
)
@option("--out", type=C_Path(dir_okay=True), required=True)
@option("--clarity", is_flag=True)
def export(collection_path, max_runs, out, clarity):
    saved_db_path = db_path(collection_path)
    if not saved_db_path.exists():
        raise FileNotFoundError(f"DB file not in {collection_path}")

    resolved = Path(out).resolve()
    resolved.mkdir(exist_ok=True)

    logger.debug(
        "we are currently at: {}. assuming npm in scope, finding dashboard lib.",
        Path.cwd(),
    )

    node_modules = check_output(
        "npm root", shell=True, text=True, cwd=Path.cwd()
    ).strip()

    logger.info("Found Node modules at: {}", node_modules)

    handshake = Path(node_modules) / "handshake-dashboard"
    if not handshake.exists():
        logger.error(
            "handshake-dashboard was not found in packages located at {}. please try to install the dashboard through "
            "the command: npm install handshake",
            handshake,
        )
        raise FileNotFoundError(
            "Please install handshake-dashboard in your project, npm install handshake"
        )

    logger.debug("Given details are valid, creating a export ticket")

    ticket_i_ds = []
    runs = []
    run_async(createExportTicket(max_runs, saved_db_path, ticket_i_ds, runs, clarity))

    if len(ticket_i_ds) == 0:
        logger.error("Ticket was not created, please report this as a issue.")

    logger.info("Exporting results to {}", relpath(resolved, handshake))
    logger.debug(
        "building files, with command:"
        ' "npx cross-env TICKET_ID={} EXPORT_DIR={}'
        ' DB_PATH={} npm run export"',
        ticket_i_ds[0],
        relpath(resolved, handshake),
        relpath(saved_db_path, handshake),
    )

    exported = call(
        f"npx cross-env TICKET_ID={ticket_i_ds[0]} EXPORT_DIR={relpath(resolved, handshake)}"
        f" DB_PATH={relpath(saved_db_path, handshake)} CLARITY={getenv('CLARITY')} npm run export",
        cwd=handshake,
        shell=True,
    )

    if ticket_i_ds:
        run_async(deleteExportTicket(ticket_i_ds[0]))

    attachments = Path(out) / writtenAttachmentFolderName
    attachments.mkdir(exist_ok=True)

    attachments_from = Path(collection_path) / writtenAttachmentFolderName

    if not (Path(out).exists() and exported == 0 and attachments_from.exists()):
        return logger.warning("Task for copying attachments was skipped.")

    allocated_work = partial(
        moveTestRunsRelatedAttachment, copyFrom=attachments_from, copyTo=attachments
    )

    with ThreadPoolExecutor(max_workers=6) as workers:
        logger.info("Copying attachments for {} test runs.", len(runs))
        workers.map(allocated_work, runs)

    logger.info("Done.")


@handle_cli.command(
    short_help="Quick Config for test collection",
    help="configures few values which is used while processing your reports., ignore the options if not required for "
    "update",
)
@general_but_optional_requirement
@option(
    "--max_runs",
    "-mr",
    default=-1,
    help="Max. Number of runs to keep. NOTE: should be >1",
)
def config(collection_path, max_runs):
    saved_db_path = db_path(collection_path)

    set_default_first = not Path(collection_path).exists()
    if set_default_first:
        Path(collection_path).mkdir(exist_ok=True)

    feed_from = config_file(saved_db_path)
    feed = dict() if not feed_from.exists() else json.loads(feed_from.read_text())
    if max_runs > 1:
        feed[ConfigKeys.maxRuns] = max_runs

    run_async(setConfig(saved_db_path, feed, set_default_first))


@handle_cli.group(
    name="db",
    short_help="Commands that will/(try to) fetch mostly asked info. from db",
    help="you can query using db following the subcommands, which are created to provide mostly asked info. from db",
)
@version_option(DB_VERSION, prog_name="handshake-db")
@general_requirement
def db(collection_path: str):
    if not Path(db_path(collection_path)).exists():
        raise FileNotFoundError(
            f"db path was not found in the collections: {collection_path}"
        )


@db.command(short_help="fetches the timestamp of the latest run")
@option(
    "-p",
    "--allow-pending",
    default=False,
    show_default=True,
    help="consider runs, whose status are pending",
    type=bool,
    is_flag=True,
)
@pass_context
def latest_run(ctx: Context, allow_pending: bool):
    db_file = db_path(Path(ctx.parent.params["collection_path"]))
    pipe = connect(db_file)
    result = pipe.execute(
        "SELECT ENDED FROM RUNBASE WHERE ENDED <> '' ORDER BY STARTED LIMIT 1"
        if not allow_pending
        else "SELECT ENDED FROM RUNBASE ORDER BY STARTED LIMIT 1"
    ).fetchone()

    secho(
        "No Test Runs were found"
        if not result
        else datetime.datetime.fromisoformat(result[0]).astimezone().strftime("%c %Z"),
        fg="bright_yellow" if not result else "bright_magenta",
    )

    pipe.close()


@db.command(
    short_help="fetches the number of yet to patch task",
    help="returns list of tasks of form: (ticket_id, task_type, dropped_date, is_picked, test_id)",
)
@pass_context
def yet_to_process(ctx: Context):
    db_file = db_path(Path(ctx.parent.params["collection_path"]))
    pipe = connect(db_file)
    result = pipe.execute(
        "SELECT ticketID, type, STRFTIME('%d/%m/%Y, %H:%M', dropped, 'localtime'), picked, test_id FROM TASKBASE WHERE "
        "PROCESSED = 0"
    ).fetchall()

    secho(
        "No Pending Tasks"
        if not result
        else f"pending tasks:\n {pprint.pformat(result)}",
        fg="bright_green" if not result else "bright_yellow",
    )

    pipe.close()



from loguru import logger
from handshake.services.DBService.models.dynamic_base import TaskBase, JobType
from uuid import uuid4, UUID
from typing import Union
from handshake.services.DBService.models.result_base import TestLogBase, LogType
from handshake.services.DBService.lifecycle import init_tortoise_orm, close_connection
from handshake.services.DBService.models.config_base import ExportBase
from handshake.services.DBService.models.result_base import RunBase
from pathlib import Path
from typing import List
from os import getenv


async def register_patch_suite(suiteID: str, testID: str) -> TaskBase:
    return await TaskBase.create(
        ticketID=suiteID, test_id=testID, type=JobType.MODIFY_SUITE
    )


async def register_patch_test_run(testID: str) -> TaskBase:
    return await TaskBase.create(
        type=JobType.MODIFY_TEST_RUN, test_id=testID, ticketID=testID
    )


async def mark_for_prune_task(test_id: str):
    # someone called this explicitly hence it's a warning

    logger.warning("Requested to prune some tasks")
    await TaskBase.create(
        ticketID=str(uuid4()), type=JobType.PRUNE_TASKS, test_id=test_id
    )


async def skip_test_run(test_id: Union[str, UUID], reason: str, **extra) -> False:
    logger.error(reason)
    await TestLogBase.create(
        test_id=str(test_id), message=reason, type=LogType.ERROR, feed=extra
    )
    await mark_for_prune_task(test_id)
    return False


async def warn_about_test_run(test_id: Union[str, UUID], about: str, **extra) -> True:
    logger.warning(about)
    await TestLogBase.create(
        test_id=str(test_id), message=about, type=LogType.WARN, feed=extra
    )
    return True


async def createExportTicket(
    maxTestRuns: int, path: Path, store: List[str], runs: List[str], clarity
):
    await init_tortoise_orm(path)
    ticket = await ExportBase.create(maxTestRuns=maxTestRuns)

    runs.extend(
        await RunBase.all()
        .order_by("-started")
        .limit(maxTestRuns)
        .values_list("testID", flat=True)
    )

    await close_connection()
    store.append(str(ticket.ticketID))


async def deleteExportTicket(ticketID: str):
    task = await ExportBase.filter(ticketID=ticketID).first()
    logger.info("Deleting the Export ticket: {}", ticketID)
    await task.delete()



from handshake.services.DBService.models.result_base import RunBase
from handshake.services.DBService.models.config_base import ConfigBase, ConfigKeys
from handshake.services.SchedularService.constants import (
    JobType,
)
from loguru import logger
from datetime import datetime
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from pathlib import Path
from typing import List
from tortoise.expressions import Q
from handshake.services.SchedularService.handleTestResults import (
    deleteTestRunsRelatedAttachments,
)


def addDeleteJob(_scheduler: AsyncIOScheduler, path: Path, mapped: List[bool]):
    _scheduler.add_job(
        deleteOldRuns,
        id=JobType.DELETE_RUNS,
        name="clearing up the old runs",
        args=(path, mapped),
        next_run_time=datetime.now(),
        max_instances=2,
    )


async def deleteOldRuns(db_path: Path, punch_out: List[bool]):
    max_requested = await ConfigBase.filter(key=ConfigKeys.maxRuns).first()
    count_of_all_runs = await RunBase.all().count()

    requested = max(int(max_requested.value), 2)

    if count_of_all_runs <= requested:
        logger.info(
            "Limit not exceeded, hence skipping deleteRuns Job. Found {} but max is {}",
            count_of_all_runs,
            requested,
        )
        return punch_out.pop()

    recently_deleted = count_of_all_runs - requested

    runs = (
        await RunBase.all()
        .order_by("started")
        .limit(count_of_all_runs - requested)
        .values_list("testID", flat=True)
    )
    await RunBase.filter(Q(testID__in=runs)).delete()

    deleteTestRunsRelatedAttachments(db_path, runs)

    record = await ConfigBase.filter(key=ConfigKeys.recentlyDeleted).first()
    if record:
        await record.update_from_dict(dict(value=str(recently_deleted)))
        await record.save()
    else:
        await ConfigBase.create(
            key=ConfigKeys.recentlyDeleted, value=str(recently_deleted)
        )

    logger.info("Delete job is completed.")
    return punch_out.pop()



from typing import TypedDict, List, Dict, Union

Node = TypedDict(
    "Node", {
        "<path>": str
    }, total=False
)
PathTree = Dict[str, Union[Node, str]]


class PathItem(TypedDict):
    pointer: PathTree
    children: List[str]



from enum import StrEnum


class JobType(StrEnum):
    INIT_CONNECTION_JOBS = "set-db-connection-add-jobs"
    LOOKUP_JOB = "pending-tasks"
    MODIFY_SUITE = "fix-suite"
    MODIFY_TEST_RUN = "fix-test-run"
    EXECUTOR = "say-bye-if-required"
    PRUNE_TASKS = "prune-tasks"
    DELETE_RUNS = "delete-runs"


writtenAttachmentFolderName = "Attachments"



from handshake.services.SchedularService.constants import JobType
from handshake.services.SchedularService.handlePending import add_lookup_task
from handshake.services.DBService.lifecycle import init_tortoise_orm
from handshake.services.SchedularService.lifecycle import verify_pending_jobs
from handshake.services.SchedularService.pruneTasks import pruneTasks
from handshake.services.SchedularService.deleteRuns import addDeleteJob
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from loguru import logger
from datetime import datetime
from typing import List
from handshake.services.DBService.models.dynamic_base import TaskBase
from handshake.services.DBService.models.config_base import ConfigBase, ConfigKeys
from handshake.services.DBService.models.result_base import RunBase, Status
from apscheduler.events import EVENT_JOB_EXECUTED
from pathlib import Path


def start_service(db_path: Path, reset: bool) -> AsyncIOScheduler:
    logger.info("Starting the scheduler service...")
    __scheduler = AsyncIOScheduler({"logger": logger})

    # for deletes
    mapped = [False]

    __scheduler.add_job(
        init_jobs_connections,
        id=JobType.INIT_CONNECTION_JOBS,
        args=(db_path, __scheduler, mapped, reset),
        next_run_time=datetime.now(),
    )
    # safe function #1
    # scheduler closes itself whenever it's jobs are done

    __scheduler.add_listener(
        lambda _: verify_pending_jobs(__scheduler, mapped), EVENT_JOB_EXECUTED
    )
    __scheduler.start()
    return __scheduler


async def pick_previous_tasks(reset: bool = False):
    await pruneTasks()
    # since we are starting a-new we would just pick the ones which are picked previously
    prev_picked_tasks = await TaskBase.filter(processed=False, picked=True).all()
    reset_from_config = await ConfigBase.filter(key=ConfigKeys.reset_test_run).first()

    reset_test_run = reset or reset_from_config.value

    to_modify_test_runs = (
        await TaskBase.filter(type=JobType.MODIFY_TEST_RUN, processed=True).all()
        if reset_test_run
        else []
    )

    to_pick = prev_picked_tasks + to_modify_test_runs
    for task in to_pick:
        logger.info("scheduling old task {} for this iteration", task.ticketID)
        task.picked = False
        task.processed = False

    if to_pick:
        await TaskBase.bulk_update(to_pick, ("picked", "processed"), 100)

    runs = []
    for ticket in to_modify_test_runs:
        test = await ticket.test
        test.standing = "PENDING"
        runs.append(test)

    if runs:
        await RunBase.bulk_update(runs, ("standing",), 100)

    if reset_from_config:
        reset_from_config.value = ""
        await reset_from_config.save()


async def init_jobs_connections(
    db_path: Path, _scheduler: AsyncIOScheduler, mapped: List[bool], reset: bool
):
    await init_tortoise_orm(db_path, True)
    await pick_previous_tasks(reset)
    addDeleteJob(_scheduler, db_path, mapped)
    add_lookup_task(_scheduler)



from handshake.services.DBService.models.result_base import (
    SessionBase,
    RunBase,
    SuiteBase,
)
from handshake.services.DBService.models.config_base import TestConfigBase
from uuid import UUID
from typing import Union
from traceback import format_exc
from handshake.services.SchedularService.register import (
    skip_test_run,
)
from handshake.services.DBService.models.dynamic_base import TaskBase
from handshake.services.DBService.models.types import Status, SuiteType
from handshake.services.SchedularService.refer_types import PathTree, PathItem
from handshake.services.SchedularService.modifySuites import fetch_key_from_status
from tortoise.functions import Sum, Max, Min, Count, Lower
from datetime import datetime, timezone
from typing import List
from pathlib import Path
from os.path import join
from loguru import logger
from tortoise.expressions import Q


async def skip_coz_error(test_id: Union[str, UUID], reason: str, **extra) -> False:
    return await skip_test_run(test_id, reason, **extra)


async def mark_test_failure_if_required(test_id: str) -> bool:
    test_run = await RunBase.filter(testID=test_id).first()
    if test_run.ended is None:
        await skip_coz_error(
            test_id,
            f"Failed to patch the Test Run: {test_id} of project: {test_run.projectName},"
            f" because it didn't have a end date",
            incomplete=test_run.projectName,
        )
        return True

    suites = [
        str(_)
        for _ in await SuiteBase.filter(session__test_id=test_id)
        .filter(Q(standing=Status.YET_TO_CALCULATE) | Q(standing=Status.PENDING))
        .all()
        .values_list("suiteID", flat=True)
    ]

    does_it_exist = await TaskBase.exists(ticketID__in=suites)
    if does_it_exist:
        # it means there are child tasks ensured to complete the child suites
        return False

    reason = f"Failed to patch TestRun: {test_id} because of incomplete child suites"
    await skip_coz_error(test_id, reason, incomplete=suites)
    return True


def simplify_file_paths(paths: List[str]):
    tree: PathTree = {"<path>": ""}
    _paths: List[PathItem] = [
        dict(
            children=list(reversed(Path(path).parts)),
            pointer=tree,
        )
        for path in paths
    ]

    # Tree - Builder
    while _paths:
        path_to_include = _paths[-1]
        if not path_to_include["children"]:
            _paths.pop()
            continue

        parent_pointer: PathTree = path_to_include["pointer"]
        possible_parent: str = path_to_include["children"].pop()
        # that dict will be inside the tree

        pointing_to = parent_pointer.setdefault(
            possible_parent, {"<path>": join(parent_pointer["<path>"], possible_parent)}
        )
        path_to_include["pointer"] = pointing_to

    # Reducer
    while True:
        stack = [(_, tree) for _ in tree.keys() if not _.startswith("<")]
        movements = 0

        while stack:
            node_key, parent_node = stack.pop()
            target_node = parent_node[node_key]

            children = set(target_node.keys())
            children.remove("<path>")

            if len(children) > 1:
                for child_key in target_node.keys():
                    if child_key.startswith("<"):
                        continue
                    stack.append((child_key, target_node))
                continue

            if not children:
                continue
            child_key = children.pop()

            movements += 1
            target_popped = parent_node.pop(node_key)
            new_key = join(node_key, child_key)
            parent_node[new_key] = target_popped.pop(child_key)
            new_child = parent_node[new_key]
            new_child["<path>"] = join(target_popped["<path>"], child_key)

            for child_key in new_child.keys():
                if child_key.startswith("<"):
                    continue
                stack.append((child_key, new_child))

        if not movements:
            break

    return tree


async def patchValues(test_run: RunBase):
    test_id = test_run.testID

    filtered = SessionBase.filter(Q(test_id=test_id) & Q(retried=False))
    retried_sessions = await SessionBase.filter(
        Q(test_id=test_id) & Q(retried=True)
    ).count()

    actual_start = "actual_start"
    actual_end = "actual_end"

    test_result = (
        await filtered.annotate(
            passed=Sum("passed"),
            failed=Sum("failed"),
            skipped=Sum("skipped"),
            tests=Sum("tests"),
            actual_end=Max("ended"),
            actual_start=Min("started"),
        )
        .first()
        .values("passed", "failed", "skipped", "tests", actual_start, actual_end)
    )

    test_config = await TestConfigBase.filter(test_id=test_id).first()

    refer = SuiteBase

    if test_config:
        # consider cucumber files
        # if the feature file has 3 scenarios, then if user sets the avoidParentSuitesInCount
        # we would get a value of 3 as total scenarios else 4
        if test_config.avoidParentSuitesInCount:
            refer = SuiteBase.filter(~Q(parent=""))

    # we want to count the number of suites status
    summary = dict(passed=0, failed=0, skipped=0)
    summary.update(
        await refer.filter(
            Q(session__test_id=test_id)
            & Q(suiteType=SuiteType.SUITE)
            & ~Q(standing=Status.RETRIED)
        )
        .annotate(count=Count("suiteID"), status=Lower("standing"))
        .group_by("standing")
        .values_list("status", "count")
    )
    summary["count"] = sum(summary.values())

    # start date was initially when we start the shipment
    # now it is when the first session starts
    started = test_result.get(actual_start, test_run.started) or test_run.started
    ended = test_result.get(actual_end, datetime.now(timezone.utc)) or datetime.now(
        timezone.utc
    )
    test_result.pop(actual_start) if actual_start in test_result else ...
    test_result.pop(actual_end) if actual_end in test_result else ...

    await test_run.update_from_dict(
        dict(
            **{_: test_result[_] or 0 for _ in test_result.keys()},
            retried=retried_sessions,
            started=started,
            ended=ended,
            duration=(ended - started).total_seconds() * 1000,
            specStructure=simplify_file_paths(
                [
                    path
                    for path in await SuiteBase.filter(Q(session__test_id=test_id))
                    .distinct()
                    .values_list("file", flat=True)
                ]
            ),
            standing=fetch_key_from_status(
                summary["passed"], summary["failed"], summary["skipped"]
            ),
            suiteSummary=summary,
        )
    )
    return True


async def patchTestRun(test_id: str, current_test_id: str):
    test_run = await RunBase.filter(testID=test_id).first()
    if not test_run:
        return True

    task = await TaskBase.filter(ticketID=test_run.testID).first()
    if not task:
        return True

    logger.info("Patching up the Test Run: {}", test_id)

    if test_run.standing != Status.PENDING:
        logger.warning("{} was already processed", test_id)
        return True

    pending_items = (
        await SuiteBase.filter(session__test_id=test_id)
        .filter(Q(standing=Status.YET_TO_CALCULATE) | Q(standing=Status.PENDING))
        .count()
    )

    if pending_items > 0:
        if await mark_test_failure_if_required(test_id):
            return
        logger.warning(
            "Patch Test Run: {} will be picked in the next run, some of its entities are not processed yet.",
            test_id,
        )
        await task.update_from_dict(dict(picked=False))
        await task.save()  # continue in the next run
        return False

    to_return = False
    try:
        if await patchValues(test_run):
            logger.info("Completed the patch for test run: {}", test_id)
            to_return = True
    except Exception:
        await skip_coz_error(
            test_id,
            f"Failed to patch the test run, error in calculation, {format_exc()}",
        )
    finally:
        await test_run.save()

    return to_return



from handshake.services.DBService.models.dynamic_base import TaskBase
from handshake.services.SchedularService.modifySuites import patchTestSuite
from handshake.services.SchedularService.constants import JobType
from handshake.services.SchedularService.completeTestRun import patchTestRun
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from handshake.services.SchedularService.pruneTasks import pruneTasks

from loguru import logger
from tortoise.expressions import Q
from datetime import datetime


def add_lookup_task(_scheduler: AsyncIOScheduler):
    _scheduler.add_job(
        lookup_for_tasks,
        id=JobType.LOOKUP_JOB,
        name="clearing up the pending tasks if present",
        args=(_scheduler,),
        next_run_time=datetime.now(),
        max_instances=2,
    )


# NOTE: make sure to pick the task before adding a new task
async def lookup_for_tasks(_scheduler: AsyncIOScheduler):
    logger.info("Looking up for the tasks")

    task = (
        await TaskBase.filter(
            Q(picked=False)
            & Q(processed=False)
            & (
                Q(type=JobType.MODIFY_SUITE)
                | Q(type=JobType.MODIFY_TEST_RUN)
                | Q(type=JobType.PRUNE_TASKS)
            )
        )
        .order_by("dropped")
        .first()
    )  # ascending

    if not task:
        return logger.warning("No Task found in this iteration")

    await task.update_from_dict(dict(picked=True))
    await task.save()

    is_task_processed = False

    match task.type:
        case JobType.MODIFY_SUITE:
            is_task_processed = await patchTestSuite(task.ticketID, task.test_id)

        case JobType.MODIFY_TEST_RUN:
            is_task_processed = await patchTestRun(task.ticketID, task.test_id)

        case JobType.PRUNE_TASKS:
            await pruneTasks(task.ticketID)
            is_task_processed = True

        case _:
            logger.warning("Not Implemented yet..")

    if is_task_processed:
        await task.update_from_dict(dict(processed=True))
        await task.save()
    else:
        logger.info(
            "This Task: {} - {} will continue in the next iteration",
            task.ticketID,
            task.type,
        )

    logger.info("Rescheduling for lookup task")
    add_lookup_task(_scheduler)



from asyncio import get_event_loop
from typing import List
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from handshake.services.DBService.models.dynamic_base import TaskBase
from handshake.services.DBService.lifecycle import close_connection
from handshake.services.SchedularService.constants import JobType
from datetime import datetime
from loguru import logger
from asyncio import run


def verify_pending_jobs(scheduler: AsyncIOScheduler, mapped: List[bool]):
    # we are adding a job to execute an async function in sync call
    scheduler.add_job(
        say_bye_if_required,
        args=(scheduler, mapped),
        id=JobType.EXECUTOR,
        name="execute-loop-if-required",
        next_run_time=datetime.now(),
        coalesce=True,
    )


async def say_bye_if_required(scheduler: AsyncIOScheduler, mapped: List[bool]):
    pending_tasks = await TaskBase.filter(processed=False).count()

    if len(mapped) > 0 or pending_tasks > 0:
        return

    scheduler.remove_all_jobs()
    # we can't shut down scheduler because you are now inside one of the core job

    # closing db connection
    await close_connection()

    current_running_loop = get_event_loop()
    current_running_loop.stop()


def start_loop():
    loop = get_event_loop()
    try:
        loop.run_forever()  # blocking
    except (KeyboardInterrupt, SystemExit):
        logger.error("Received your signal, closing the loop.")
        run(close_connection())
    finally:
        loop.close()
